/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.optaplanner.core.impl.domain.common.accessor.gizmo;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatCode;

import java.lang.reflect.Member;
import java.lang.reflect.Method;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.optaplanner.core.api.domain.variable.PlanningVariable;
import org.optaplanner.core.impl.domain.common.accessor.MemberAccessor;
import org.optaplanner.core.impl.testdata.domain.TestdataEntity;
import org.optaplanner.core.impl.testdata.domain.TestdataValue;

class GizmoMemberAccessorFactoryTest {

    static ClassLoader contextClassLoader;

    @BeforeAll
    static void setContextClassLoader() {
        contextClassLoader = Thread.currentThread().getContextClassLoader();
    }

    @BeforeEach
    void setup() {
        Thread.currentThread().setContextClassLoader(contextClassLoader);
    }

    // Duplicates GizmoMemberAccessorImplementor test,
    // but this is making sure the member accessor returned
    // is the one from GizmoMemberAccessorImplementor
    @Test
    void testReturnedMemberAccessor() throws NoSuchMethodException {
        Method member = TestdataEntity.class.getMethod("getValue");
        MemberAccessor memberAccessor =
                GizmoMemberAccessorFactory.buildGizmoMemberAccessor(member, PlanningVariable.class, new GizmoClassLoader());

        TestdataEntity entity = new TestdataEntity();
        TestdataValue value = new TestdataValue("A");
        entity.setValue(value);
        assertThat(memberAccessor.executeGetter(entity)).isEqualTo(value);

        assertThat(memberAccessor.supportSetter()).isTrue();

        memberAccessor.executeSetter(entity, null);
        assertThat(entity.getValue()).isNull();

        assertThat(memberAccessor.getDeclaringClass()).isEqualTo(TestdataEntity.class);
        assertThat(memberAccessor.getGenericType()).isEqualTo(member.getGenericReturnType());
        assertThat(memberAccessor.getType()).isEqualTo(member.getReturnType());
        assertThat(memberAccessor.getName()).isEqualTo("value");
        assertThat(memberAccessor.getSpeedNote()).isEqualTo("Fast access with generated bytecode");
    }

    @Test
    void testGizmoNotOnClasspathThrowsException() throws NoSuchMethodException {
        Member member = TestdataEntity.class.getMethod("getValue");
        Thread.currentThread().setContextClassLoader(new ClassLoader() {
            @Override
            public String getName() {
                return "ClassLoader without Gizmo";
            }

            @Override
            public Class<?> loadClass(String name) {
                return null;
            }
        });

        assertThatCode(() -> {
            GizmoMemberAccessorFactory.buildGizmoMemberAccessor(member, PlanningVariable.class, new GizmoClassLoader());
        }).hasMessage("When using the domainAccessType (GIZMO) the classpath or modulepath must contain " +
                "io.quarkus.gizmo:gizmo.\nMaybe add a dependency to io.quarkus.gizmo:gizmo.");
    }
}
